//
// Copyright (c) 2009, Markus Rickert
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name of the Technische Universitaet Muenchen nor the names of
//   its contributors may be used to endorse or promote products derived from
//   this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//

#ifndef _RL_MATH_RIGIDBODYINERTIA_H_
#define _RL_MATH_RIGIDBODYINERTIA_H_

#include <Eigen/Core>

namespace rl
{
	namespace math
	{
		namespace spatial
		{
			template< typename Scalar > class MotionVector;
			
			template< typename Scalar >
			class RigidBodyInertia
			{
			public:
				EIGEN_MAKE_ALIGNED_OPERATOR_NEW
				
				typedef Scalar ScalarType;
				
				typedef ::Eigen::Matrix< Scalar, 6, 6 > MatrixType;
				
				typedef ::Eigen::Matrix< Scalar, 3, 1 > CenterOfGravityType;
				
				typedef const CenterOfGravityType ConstCenterOfGravityType;
				
				typedef ::Eigen::Matrix< Scalar, 3, 3 > InertiaType;
				
				typedef const InertiaType ConstInertiaType;
				
				typedef Scalar MassType;
				
				typedef const MassType ConstMassType;
				
				RigidBodyInertia()
				{
				}
				
				template< typename OtherDerived >
				RigidBodyInertia(const ::Eigen::DenseBase< OtherDerived >& other) :
					centerOfGravityData(other.template topRightCorner< 3, 3 >().cross3()),
					inertiaData(other.template topLeftCorner< 3, 3 >()),
					massData(other.template bottomRightCorner< 3, 3 >().diagonal().mean())
				{
				}
				
				virtual ~RigidBodyInertia()
				{
				}
				
				CenterOfGravityType& cog()
				{
					return centerOfGravityData;
				}
				
				ConstCenterOfGravityType& cog() const
				{
					return centerOfGravityData;
				}
				
				InertiaType& inertia()
				{
					return inertiaData;
				}
				
				ConstInertiaType& inertia() const
				{
					return inertiaData;
				}
				
				MassType& mass()
				{
					return massData;
				}
				
				ConstMassType& mass() const
				{
					return massData;
				}
				
				MatrixType matrix() const
				{
					MatrixType res;
					res.template topLeftCorner< 3, 3 >() = inertia();
					res.template topRightCorner< 3, 3 >() = cog().cross33();
					res.template bottomLeftCorner< 3, 3 >() = -cog().cross33();
					res.template bottomRightCorner< 3, 3 >() = ::Eigen::Matrix< Scalar, 3, 3 >::Identity() * mass();
					return res;
				}
				
				template< typename OtherDerived >
				RigidBodyInertia& operator=(const ::Eigen::MatrixBase< OtherDerived >& other)
				{
					cog() = other.template topRightCorner< 3, 3 >().cross3();
					inertia() = other.template topLeftCorner< 3, 3 >();
					mass() = other.template bottomRightCorner< 3, 3 >().diagonal().mean();
					return *this;
				}
				
				RigidBodyInertia operator+(const RigidBodyInertia& other) const
				{
					RigidBodyInertia res;
					res.cog() = cog() + other.cog();
					res.inertia() = inertia() + other.inertia();
					res.mass() = mass() + other.mass();
					return res;
				}
				
				RigidBodyInertia operator-(const RigidBodyInertia& other) const
				{
					RigidBodyInertia res;
					res.cog() = cog() - other.cog();
					res.inertia() = inertia() - other.inertia();
					res.mass() = mass() - other.mass();
					return res;
				}
				
				template< typename OtherScalar >
				RigidBodyInertia operator*(const OtherScalar& other) const
				{
					RigidBodyInertia res;
					res.cog() = cog() * other;
					res.inertia() = inertia() * other;
					res.mass() = mass() * other;
					return res;
				}
				
				template< typename OtherScalar >
				ForceVector< Scalar > operator*(const MotionVector< OtherScalar >& other) const;
				
				template< typename OtherScalar >
				RigidBodyInertia operator/(const OtherScalar& other) const
				{
					RigidBodyInertia res;
					res.cog() = cog() / other;
					res.inertia() = inertia() / other;
					res.mass() = mass() / other;
					return res;
				}
				
				void setIdentity()
				{
					cog().setZero();
					inertia().setIdentity();
					mass() = 1;
				}
				
				void setZero()
				{
					cog().setZero();
					inertia().setZero();
					mass() = 0;
				}
				
			protected:
				
			private:
				CenterOfGravityType centerOfGravityData;
				
				InertiaType inertiaData;
				
				MassType massData;
			};
		}
	}
}

#endif // _RL_MATH_RIGIDBODYINERTIA_H_
